@@ -17,13 +17,22 @@ import copy
 from shapely.geometry import Polygon, LineString
 from numbers import Integral
 from jsonschema import Draft7Validator, Draft4Validator, validators, ValidationError, RefResolver
-from starling_sim.utils.paths import schemas_folder, gtfs_feeds_folder, osm_graphs_folder
+from starling_sim.utils.paths import (
+    schemas_folder,
+    gtfs_feeds_folder,
+    osm_graphs_folder,
+    scenario_inputs_folder,
+    PARAMETERS_FILENAME,
+    INPUT_FOLDER_NAME,
+)
+from starling_sim.utils.simulation_logging import BLANK_LOGGER
 
-pd.set_option('display.expand_frame_repr', False)
+pd.set_option("display.expand_frame_repr", False)
 
 
 # Starling exceptions
 
+
 class StarlingException(Exception):
     """
     Base class for all Starling exceptions.
@@ -51,8 +60,16 @@ class SimulationError(LeavingSimulation):
     """
 
 
+class PlanningChange(Exception):
+    """
+    Exception raised to interrupt a service vehicle loop
+    when its planning has changed.
+    """
+
+
 # json utils
 
+
 def json_dump(data, filepath):
     """
     Create a json file and dump the given object
@@ -94,6 +111,7 @@ def json_load(filepath):
 
 # compression utils
 
+
 def gz_compression(filepath, delete_source=True):
     """
     Compress the given file using gzip.
@@ -105,14 +123,17 @@ def gz_compression(filepath, delete_source=True):
     """
 
     # compress the file using gzip
-    with open(filepath, 'rb') as f_in:
-        with gzip.open(filepath + ".gz", 'wb') as f_out:
+    compressed_path = filepath + ".gz"
+    with open(filepath, "rb") as f_in:
+        with gzip.open(compressed_path, "wb") as f_out:
             shutil.copyfileobj(f_in, f_out)
 
     # delete source file if asked
     if delete_source:
         os.remove(filepath)
 
+    return compressed_path
+
 
 def gz_decompression(filepath, delete_source=True):
     """
@@ -128,11 +149,14 @@ def gz_decompression(filepath, delete_source=True):
 
     # test if the file ends with .gz
     if not filepath.endswith(".gz"):
-        logging.log(30, "File to decompress does not end with '.gz'. Continuing without decompressing the file.")
+        logging.log(
+            30,
+            "File to decompress does not end with '.gz'. Continuing without decompressing the file.",
+        )
 
     # decompress the file unsing gzip
-    with gzip.open(filepath, 'rb') as f_in:
-        with open(filepath[:-3], 'wb') as f_out:
+    with gzip.open(filepath, "rb") as f_in:
+        with open(filepath[:-3], "wb") as f_out:
             shutil.copyfileobj(f_in, f_out)
 
     # delete source file if asked
@@ -140,6 +164,50 @@ def gz_decompression(filepath, delete_source=True):
         os.remove(filepath)
 
 
+# starling file information
+
+
+def create_file_information(
+    filepath,
+    mimetype,
+    compressed_mimetype=None,
+    content=None,
+    subject=None,
+    keep_filepath=False,
+):
+    """
+    Create a dict containing file information.
+
+    :param filepath: file path
+    :param mimetype: file mimetype
+    :param compressed_mimetype: file mimetype after decompression
+    :param content: content metadata
+    :param subject: subject metadata
+    :param keep_filepath: keep complete file path in 'filename'
+
+    :return: { filename/filepath, mimetype, metadata }
+    """
+
+    if content is None:
+        raise ValueError("'content' metadata was not provided for file {}".format(filepath))
+
+    if compressed_mimetype is None:
+        compressed_mimetype = mimetype
+
+    metadata = {"compressed-mimetype": compressed_mimetype, "content": content}
+
+    if subject is not None:
+        metadata["subject"] = subject
+
+    file_information = {
+        "filename": filepath if keep_filepath else os.path.basename(filepath),
+        "mimetype": mimetype,
+        "metadata": metadata,
+    }
+
+    return file_information
+
+
 # json schema validation
 
 # use the type check from Draft4Validator, because Draft7 considers 1.0 as integer
@@ -152,7 +220,7 @@ def validate_against_schema(instance, schema, raise_exception=True):
     schema = load_schema(schema, False)
 
     # get the absolute path and setup a resolver
-    schema_abs_path = 'file:///{0}/'.format(os.path.abspath(schemas_folder()).replace("\\", "/"))
+    schema_abs_path = "file:///{0}/".format(os.path.abspath(schemas_folder()).replace("\\", "/"))
     resolver = RefResolver(schema_abs_path, schema)
     validator = CustomValidator(schema=schema, resolver=resolver)
 
@@ -164,7 +232,12 @@ def validate_against_schema(instance, schema, raise_exception=True):
         if raise_exception:
             raise e
         else:
-            logging.log(30, "JSON Schema validation of :\n\n{}\n\nfailed with message:\n\n {} ".format(instance, e))
+            logging.log(
+                30,
+                "JSON Schema validation of :\n\n{}\n\nfailed with message:\n\n {} ".format(
+                    instance, e
+                ),
+            )
             return False
 
 
@@ -223,11 +296,12 @@ def load_schema(schema, make_copy=True):
 
 # converters
 
+
 def get_sec(time_str):
     """Get Seconds from time."""
 
     # HH:MM:SS to seconds
-    h, m, s = time_str.split(':')
+    h, m, s = time_str.split(":")
     return int(h) * 3600 + int(m) * 60 + int(s)
 
 
@@ -268,10 +342,7 @@ def new_point_feature(point_localisation=None, properties=None):
         point_localisation = [0, 0]
 
     # create  the feature geometry
-    geometry = {
-        "type": "Point",
-        "coordinates": point_localisation
-    }
+    geometry = {"type": "Point", "coordinates": point_localisation}
 
     return new_feature(geometry, properties)
 
@@ -279,10 +350,7 @@ def new_point_feature(point_localisation=None, properties=None):
 def new_line_string_feature(linestring, properties=None):
 
     # create  the feature geometry
-    geometry = {
-        "type": "LineString",
-        "coordinates": linestring
-    }
+    geometry = {"type": "LineString", "coordinates": linestring}
 
     return new_feature(geometry, properties)
 
@@ -297,10 +365,7 @@ def new_multi_polygon_feature(polygon_list, properties=None):
     """
 
     # create  the feature geometry
-    geometry = {
-        "type": "MultiPolygon",
-        "coordinates": polygon_list
-    }
+    geometry = {"type": "MultiPolygon", "coordinates": polygon_list}
 
     return new_feature(geometry, properties)
 
@@ -312,11 +377,7 @@ def new_feature(geometry, properties=None):
         properties = dict()
 
     # build a geojson feature
-    feature = {
-        "type": "Feature",
-        "geometry": geometry,
-        "properties": properties
-    }
+    feature = {"type": "Feature", "geometry": geometry, "properties": properties}
 
     return feature
 
@@ -333,16 +394,14 @@ def new_feature_collection(features=None):
         features = []
 
     # build a geojson feature collection
-    feature_collection = {
-        "type": "FeatureCollection",
-        "features": features
-    }
+    feature_collection = {"type": "FeatureCollection", "features": features}
 
     return feature_collection
 
 
 # shapes functions
 
+
 def shapely_polygon_from_points(points, reverse=False):
     """
     Create and return a Shapely Polygon object.
@@ -386,12 +445,14 @@ def geopandas_points_from_localisations(localisations):
 
     if isinstance(localisations, tuple) or isinstance(localisations, list):
         # create a DataFrame containing the point coordinates
-        localisations = pd.DataFrame({"lat": localisations[0], "lon": localisations[1]},
-                                     columns=["lat", "lon"], index=[0])
+        localisations = pd.DataFrame(
+            {"lat": localisations[0], "lon": localisations[1]}, columns=["lat", "lon"], index=[0]
+        )
 
     # create a GeoDataFrame from df
     gdf = geopandas.GeoDataFrame(
-        localisations, geometry=geopandas.points_from_xy(localisations.lon, localisations.lat))
+        localisations, geometry=geopandas.points_from_xy(localisations.lon, localisations.lat)
+    )
 
     # set epsg
     gdf.set_crs("epsg:4326")
@@ -498,20 +559,27 @@ def stops_table_from_geojson(geojson_path):
         if "stop_name" not in properties:
             properties["stop_name"] = "Stop point {id}".format(id=properties["stop_id"])
 
+        # convert properties to DataFrame
+        properties = pd.DataFrame([properties])
+
         # append a new row
-        stops_table = stops_table.append(properties, ignore_index=True)
+        stops_table = pd.concat([stops_table, properties], ignore_index=True)
 
     return stops_table
 
 
-def stop_table_from_gtfs(gtfs_feed, routes=None, zone=None, fixed_stops=None, active_stops_only=False):
+def stop_table_from_gtfs(
+    gtfs_feed, routes=None, zone=None, fixed_stops=None, active_stops_only=False
+):
 
     result_table = pd.DataFrame()
 
     gtfs_stops = gtfs_feed.get_stops().copy()
 
     if fixed_stops is not None:
-        result_table = pd.concat([result_table, gtfs_stops[gtfs_stops["stop_id"].isin(fixed_stops)]], sort=False)
+        result_table = pd.concat(
+            [result_table, gtfs_stops[gtfs_stops["stop_id"].isin(fixed_stops)]], sort=False
+        )
 
     if active_stops_only:
         stop_times = gtfs_feed.get_stop_times()
@@ -529,7 +597,9 @@ def stop_table_from_gtfs(gtfs_feed, routes=None, zone=None, fixed_stops=None, ac
         gtfs_stops = gtfs_stops[gtfs_stops["in_zone"]]
 
         # remove stops areas
-        gtfs_stops = gtfs_stops[(gtfs_stops["location_type"] == 0) | (gtfs_stops["location_type"].isna())]
+        gtfs_stops = gtfs_stops[
+            (gtfs_stops["location_type"] == 0) | (gtfs_stops["location_type"].isna())
+        ]
 
     result_table = pd.concat([result_table, gtfs_stops], sort=False)
     result_table.drop_duplicates(inplace=True)
@@ -537,8 +607,17 @@ def stop_table_from_gtfs(gtfs_feed, routes=None, zone=None, fixed_stops=None, ac
     return result_table
 
 
-def import_osm_graph(method, network_type, simplify, query=None, which_result=None, point=None, dist=None, polygon=None,
-                     outfile=None):
+def import_osm_graph(
+    method,
+    network_type,
+    simplify,
+    query=None,
+    which_result=None,
+    point=None,
+    dist=None,
+    polygon=None,
+    outfile=None,
+):
     """
     Generate an OSM graph from given parameters and store it in a file.
 
@@ -612,14 +691,17 @@ def osm_graph_from_point(point, distance, network_type, simplify):
     """
 
     if point is None or distance is None:
-        print("The point and distance parameters must be specified when importing graph from point.")
+        print(
+            "The point and distance parameters must be specified when importing graph from point."
+        )
         exit(1)
 
     # reverse the point coordinates (osmnx takes (lat, lon) coordinates)
     point = (point[1], point[0])
 
-    return ox.graph_from_point(point, dist=distance, dist_type="bbox",
-                               network_type=network_type, simplify=simplify)
+    return ox.graph_from_point(
+        point, dist=distance, dist_type="bbox", network_type=network_type, simplify=simplify
+    )
 
 
 def osm_graph_from_polygon(polygon_points, network_type, simplify):
@@ -660,7 +742,9 @@ def osm_graph_from_place(query, which_result, network_type, simplify):
         print("The query parameter must be specified when importing graph from place.")
         exit(1)
 
-    return ox.graph_from_place(query, network_type=network_type, simplify=simplify, which_result=which_result)
+    return ox.graph_from_place(
+        query, network_type=network_type, simplify=simplify, which_result=which_result
+    )
 
 
 def save_osm_graph(graph, filename, folder):
@@ -719,6 +803,7 @@ def osm_graph_from_file(filename, folder=None):
 
 # gtfs utils
 
+
 def import_gtfs_feed(gtfs_filename, transfer_restriction=None, folder=None):
     """
     Import a gtfs feed from the given file.
@@ -751,14 +836,19 @@ def import_gtfs_feed(gtfs_filename, transfer_restriction=None, folder=None):
     # check that foot-path transfers are symmetrical
     if feed.transfers is not None:
         transfer_table = feed.transfers.copy()
-        transfer_table = transfer_table[transfer_table["from_stop_id"] != transfer_table["to_stop_id"]]
+        transfer_table = transfer_table[
+            transfer_table["from_stop_id"] != transfer_table["to_stop_id"]
+        ]
         transfer_table["stop_A"] = transfer_table[["from_stop_id", "to_stop_id"]].apply(min, axis=1)
         transfer_table["stop_B"] = transfer_table[["from_stop_id", "to_stop_id"]].apply(max, axis=1)
         count = transfer_table.groupby(["stop_A", "stop_B"], as_index=False).agg(["count"])
         counts_not_equal_to_2 = count[count["min_transfer_time"]["count"] != 2]
         if not counts_not_equal_to_2.empty:
-            logging.warning("Transfer table of {} is not symmetrical (in term of arcs, not transfer times)"
-                            .format(gtfs_filename))
+            logging.warning(
+                "Transfer table of {} is not symmetrical (in term of arcs, not transfer times)".format(
+                    gtfs_filename
+                )
+            )
         if not is_transitive(feed.transfers) and transfer_restriction is not None:
             feed.transfers = transitively_closed_transfers(feed.transfers, transfer_restriction)
     else:
@@ -782,9 +872,12 @@ def transitively_closed_transfers(transfers, restrict_transfer_time):
 
     # restrict original transfers so the final set of transfers isn't too large
     if restrict_transfer_time is not None:
-        logging.info("The GTFS transfer table is not transitively closed. "
-                     "Transfers are restricted to duration under {} seconds and then made transitive."
-                     .format(restrict_transfer_time))
+        logging.info(
+            "The GTFS transfer table is not transitively closed. "
+            "Transfers are restricted to duration under {} seconds and then made transitive.".format(
+                restrict_transfer_time
+            )
+        )
         transfers = transfers[transfers["min_transfer_time"] <= restrict_transfer_time]
         return make_transfers_transitively_closed(transfers)
     # else:
@@ -819,11 +912,15 @@ def is_transitive(transfers):
     nb_transfers = len(transfers)
     # create new transfers using transitive property
     new_transfers = pd.merge(transfers, transfers, left_on="to_stop_id", right_on="from_stop_id")
-    new_transfers["min_transfer_time"] = new_transfers["min_transfer_time_x"] + new_transfers["min_transfer_time_y"]
+    new_transfers["min_transfer_time"] = (
+        new_transfers["min_transfer_time_x"] + new_transfers["min_transfer_time_y"]
+    )
     new_transfers["from_stop_id"] = new_transfers["from_stop_id_x"]
     new_transfers["to_stop_id"] = new_transfers["to_stop_id_y"]
     if "transfer_type" in transfers.columns:
-        new_transfers["transfer_type"] = new_transfers[["transfer_type_x", "transfer_type_y"]].max(axis=1)
+        new_transfers["transfer_type"] = new_transfers[["transfer_type_x", "transfer_type_y"]].max(
+            axis=1
+        )
     new_transfers = new_transfers[transfers.columns]
 
     # add them to the former set of transfers
@@ -854,12 +951,18 @@ def make_transfers_transitively_closed(transfers):
     nb_transfers = len(transfers)
     while True:
         # create new transfers using transitive property
-        new_transfers = pd.merge(transfers, transfers, left_on="to_stop_id", right_on="from_stop_id")
-        new_transfers["min_transfer_time"] = new_transfers["min_transfer_time_x"] + new_transfers["min_transfer_time_y"]
+        new_transfers = pd.merge(
+            transfers, transfers, left_on="to_stop_id", right_on="from_stop_id"
+        )
+        new_transfers["min_transfer_time"] = (
+            new_transfers["min_transfer_time_x"] + new_transfers["min_transfer_time_y"]
+        )
         new_transfers["from_stop_id"] = new_transfers["from_stop_id_x"]
         new_transfers["to_stop_id"] = new_transfers["to_stop_id_y"]
         if "transfer_type" in transfers.columns:
-            new_transfers["transfer_type"] = new_transfers[["transfer_type_x", "transfer_type_y"]].max(axis=1)
+            new_transfers["transfer_type"] = new_transfers[
+                ["transfer_type_x", "transfer_type_y"]
+            ].max(axis=1)
         new_transfers = new_transfers[transfers.columns]
 
         # add them to the former set of transfers
@@ -888,12 +991,18 @@ def stops2geojson(stops_df):
     """
 
     features = []
-    stops_df.apply(lambda x: features.append(
-        new_point_feature(point_localisation=[x["stop_lon"], x["stop_lat"]],
-                          properties={
-                              "stop_id": x["stop_id"],
-                              "stop_name": x["stop_name"],
-                          })), axis=1)
+    stops_df.apply(
+        lambda x: features.append(
+            new_point_feature(
+                point_localisation=[x["stop_lon"], x["stop_lat"]],
+                properties={
+                    "stop_id": x["stop_id"],
+                    "stop_name": x["stop_name"],
+                },
+            )
+        ),
+        axis=1,
+    )
 
     return new_feature_collection(features)
 
@@ -911,12 +1020,25 @@ def transfers2geojson(transfers_df, stops_df):
     # add the stop information to the from_stop and to_stop
     geo_transfers = pd.merge(transfers_df, stops_df, left_on=["from_stop_id"], right_on=["stop_id"])
     geo_transfers = pd.merge(geo_transfers, stops_df, left_on=["to_stop_id"], right_on=["stop_id"])
-    geo_transfers = geo_transfers[["from_stop_id", "to_stop_id", "min_transfer_time", "stop_lat_x", "stop_lon_x",
-                                   "stop_lat_y", "stop_lon_y"]]
+    geo_transfers = geo_transfers[
+        [
+            "from_stop_id",
+            "to_stop_id",
+            "min_transfer_time",
+            "stop_lat_x",
+            "stop_lon_x",
+            "stop_lat_y",
+            "stop_lon_y",
+        ]
+    ]
 
     # create the geometry attribute
     geo_transfers["geometry"] = geo_transfers.apply(
-        lambda x: LineString([(x["stop_lon_x"], x["stop_lat_x"]), (x["stop_lon_y"], x["stop_lat_y"])]), axis=1)
+        lambda x: LineString(
+            [(x["stop_lon_x"], x["stop_lat_x"]), (x["stop_lon_y"], x["stop_lat_y"])]
+        ),
+        axis=1,
+    )
     geo_transfers = geopandas.GeoDataFrame(geo_transfers)
     geo_transfers = geo_transfers[["min_transfer_time", "geometry"]]
 
@@ -927,6 +1049,7 @@ def transfers2geojson(transfers_df, stops_df):
 
 # folder creation
 
+
 def create_if_not_exists(folder):
     """
     Test the existence of the folder and create it if not present.
@@ -939,3 +1062,80 @@ def create_if_not_exists(folder):
         return True
     else:
         return False
+
+
+# console log
+
+
+def display_horizontal_bar(lvl=20):
+    """
+    Display a horizontal bar in the terminal.
+
+    Try to make it the size of the terminal width.
+
+    :param lvl: log level
+    """
+
+    try:
+        bar_size = os.get_terminal_size().columns
+    except OSError:
+        bar_size = 100
+
+    BLANK_LOGGER.log(lvl, "\u2014" * bar_size)
+
+
+# multiple scenarios
+
+
+def create_sub_scenarios(simulation_scenario):
+
+    logging.info("Creating sub scenarios of: " + simulation_scenario.name)
+
+    nb_scenarios = simulation_scenario["multiple"]
+    scenarios_folder = os.path.join(simulation_scenario.scenario_folder, "scenarios")
+    create_if_not_exists(scenarios_folder)
+
+    # set random seed
+    np.random.seed(simulation_scenario["seed"])
+    seeds = np.random.choice(range(100000), size=nb_scenarios, replace=False)
+    sub_scenario_name_format = "{base_scenario}-{index}"
+
+    scenario_paths = []
+    for i in range(nb_scenarios):
+
+        sub_scenario_name = sub_scenario_name_format.format(
+            base_scenario=simulation_scenario.name, index=i
+        )
+        sub_scenario_folder = os.path.join(scenarios_folder, sub_scenario_name)
+        sub_scenario_inputs_folder = scenario_inputs_folder(sub_scenario_folder)
+        scenario_paths.append(sub_scenario_folder)
+
+        if os.path.exists(sub_scenario_folder):
+            logging.info("Scenario {} already created".format(sub_scenario_name))
+            continue
+        else:
+            logging.info("Creating scenario " + sub_scenario_name)
+
+        # create sub scenario folders
+        create_if_not_exists(sub_scenario_folder)
+        create_if_not_exists(sub_scenario_inputs_folder)
+
+        # sub scenario inputs
+        sub_parameters = simulation_scenario.copy_parameters()
+        sub_parameters["scenario"] = sub_scenario_name
+        sub_parameters["seed"] = int(seeds[i])
+        del sub_parameters["multiple"]
+        json_pretty_dump(
+            sub_parameters, os.path.join(sub_scenario_inputs_folder, PARAMETERS_FILENAME)
+        )
+
+        # create symlinks to original inputs
+        for input_file in os.listdir(simulation_scenario.inputs_folder):
+            if input_file == PARAMETERS_FILENAME:
+                continue
+            input_filepath = os.path.join("..", "..", "..", INPUT_FOLDER_NAME, input_file)
+            os.symlink(input_filepath, os.path.join(sub_scenario_inputs_folder, input_file))
+
+    logging.info("End of sub scenarios creation\n")
+
+    return scenario_paths
